/*******************************************************************************
 *	Maps.js		Websys maps javascript
 *	T.Barnaby,	Beamweb,	2019-04-19
 *	Copyright 2021 BEAM Ltd.
 *******************************************************************************
 */

/*	Copyright (c) 2016 Jean-Marc VIGLINO, 
  released under the CeCILL-B license (French BSD license)
  (http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.txt).
*/
/*eslint no-constant-condition: ["error", { "checkLoops": false }]*/

function tableAddParam(row, name, title){
	var th = document.createElement("th");
	th.innerHTML = title;
	var td = document.createElement("td");
	td.innerHTML = "<span class='info_" + name + "'>";
	row.appendChild(th);
	row.appendChild(td);
}

/**
 * @classdesc OpenLayers 3 Profil Control.
 * Draw a profil of a feature (with a 3D geometry)
 *
 * @constructor
 * @extends {ol.control.Control}
 * @fires  over, out, show
 * @param {Object=} options
 *  @param {string} className
 *  @param {ol.style.Style} style style to draw the profil
 *  @param {*} info keys/values for i19n
 *  @param {number} width
 *  @param {number} height
 *  @parma {ol.Feature} feature the feature to draw profil
 */
ol.control.Profil_beam = function(opt_options) {
  var options = opt_options || {};
  this.info = options.info || ol.control.Profil_beam.prototype.info;
  var self = this;
  var element;
  if (options.target) {
    element = document.createElement("div");
    element.classList.add(options.className || "ol-profil");
  } else {
    element = document.createElement("div");
    element.className = ((options.className || 'ol-profil') +' ol-unselectable ol-control ol-collapsed').trim();
    this.button = document.createElement("button");
    this.button.setAttribute('type','button');
    var click_touchstart_function = function(e) {
      self.toggle();
      e.preventDefault();
    };
    this.button.addEventListener("click", click_touchstart_function);
    this.button.addEventListener("touchstart", click_touchstart_function);
    element.appendChild(this.button);
  }
  if (options.style instanceof ol.style.Style) {
    this._style = options.style;
  } else {
    this._style = new ol.style.Style({
      text: new ol.style.Text(),
      stroke: new ol.style.Stroke({
        width: 1.5,
        color: '#369'
      })
    });
  }
  if (!this._style.getText()) this._style.setText(new ol.style.Text());
  var div_inner = document.createElement("div");
      div_inner.classList.add("ol-inner");
      element.appendChild(div_inner);
  var div = document.createElement("div");
      //div.style.position = "relative";
      div_inner.appendChild(div);
  var ratio = this.ratio = 1;
  this.canvas_ = document.createElement('canvas');
  this.canvas_.width = (options.width || 300)*ratio;
  this.canvas_.height = (options.height || 150)*ratio;
  this.canvas_.style.width = "100%";

  var div_to_canvas = document.createElement("div");
  div.appendChild(div_to_canvas);
  div_to_canvas.style.width = "100%";
  div_to_canvas.style.height = this.canvas_.height/ratio + "px";

  div_to_canvas.appendChild(this.canvas_);
  div_to_canvas.addEventListener("click", function(e){ self.onMove(e); });
  div_to_canvas.addEventListener("mousemove", function(e){ self.onMove(e); });
  ol.control.Control.call(this, {
    element: element,
    target: options.target
  });
  // Offset in px
  this.margin_ = { top:10*ratio, left:40*ratio, bottom:30*ratio, right:10*ratio };
  if (!this.info.ytitle) this.margin_.left -= 20*ratio;
  if (!this.info.xtitle) this.margin_.bottom -= 20*ratio;
  // Cursor
  this.bar_ = document.createElement("div");
  this.bar_.classList.add("ol-profilbar");
  this.bar_.style.top = (this.margin_.top/ratio)+"px";
  this.bar_.style.height = (this.canvas_.height-this.margin_.top-this.margin_.bottom)/ratio+"px";
  div.appendChild(this.bar_);
  this.cursor_ = document.createElement("div");
  this.cursor_.classList.add("ol-profilcursor");
  div.appendChild(this.cursor_);
  this.popup_ = document.createElement("div");
  this.popup_.classList.add("ol-profilpopup");
  this.cursor_.appendChild(this.popup_);
  // Track information
  var t = document.createElement("table");
      t.cellPadding = '0';
      t.cellSpacing = '0';
      //t.style.width = "100%";
    div.appendChild(t);

  var firstTr = document.createElement("tr"); firstTr.classList.add("track-info"); t.appendChild(firstTr);
  tableAddParam(firstTr, "totalDistance", "TotalDistance:");
  tableAddParam(firstTr, "climb", "Climb:");
  tableAddParam(firstTr, "effort", "Effort:");
  
  var secondTr = document.createElement("tr"); secondTr.classList.add("point-info"); t.appendChild(secondTr);
  tableAddParam(secondTr, "altitude", "Altitude:");
  tableAddParam(secondTr, "distance", "Distance:");
  tableAddParam(secondTr, "gradient", "Gradient:");
      
  // Array of data
  this.tab_ = [];
  // Show feature
  if (options.feature) {
    this.setGeometry (options.feature);
  }
};

ol.ext.inherits(ol.control.Profil_beam, ol.control.Control);

/** Custom infos list
* @api stable
*/
ol.control.Profil_beam.prototype.info = {
  "zmin": "Zmin",
  "zmax": "Zmax",
  "ytitle": "Altitude (m)",
  "xtitle": "Distance (km)",
  "time": "Time",
  "altitude": "Altitude",
  "distance": "Distance",
  "altitudeUnits": "m",
  "distanceUnitsM": "m",
  "distanceUnitsKM": "km",
};
/** Show popup info
* @param {string} info to display as a popup
* @api stable
*/
ol.control.Profil_beam.prototype.popup = function(info) {
  this.popup_.innerHTML = info;
};
/** Show point on profil
 * @param {*} p 
 * @param {number} dx 
 * @private
 */
ol.control.Profil_beam.prototype._drawAt = function(p, dx) {
  if (p) {
    this.cursor_.style.left = dx+"px";
    this.cursor_.style.top = (this.canvas_.height-this.margin_.bottom+p[1]*this.scale_[1]+this.dy_)/this.ratio+"px";
    this.cursor_.style.display = "block";
    this.bar_.parentElement.classList.add("over");
    this.bar_.style.left = dx+"px";
    this.bar_.style.display = "block";
    this.element.querySelector(".info_altitude").textContent = p[1]+this.info.altitudeUnits;
    this.element.querySelector(".info_distance").textContent = (p[0]/1000).toFixed(1)+this.info.distanceUnitsKM;
    this.element.querySelector(".info_gradient").textContent = (100.0 * p[4]).toFixed(1) + "%";

    //this.element.querySelector(".point-info .time").textContent = p[2];
    if (dx>this.canvas_.width/this.ratio/2) this.popup_.classList.add('ol-left');
    else this.popup_.classList.remove('ol-left');
  } else {
    this.cursor_.style.display = "none";
    this.bar_.style.display = 'none';
    this.cursor_.style.display = 'none';  
    this.bar_.parentElement.classList.remove("over");
  }
};
/** Show point at coordinate or a distance on the profil
 * @param { ol.coordinates||number } where a coordiniate or a distance from begining, if none it will hide the point
 * @return { ol.coordinates } current point
 */
ol.control.Profil_beam.prototype.showAt = function(where) {
  var i, p, p0, d0 = Infinity;
  if (typeof(where) === 'undefined') {
    if (this.bar_.parentElement.classList.contains("over")) {
      // Remove it
      this._drawAt();
    }
  } else if (where.length) {
    // Look for closest the point
    for (i=1; p=this.tab_[i]; i++) {
      var d = ol.coordinate.dist2d(p[3], where);
      if (d<d0) {
        p0 = p;
        d0 = d;
      } 
    }
  } else {
    for (i=0; p=this.tab_[i]; i++) {
      p0 = p;
      if (p[0] > where) {
        break;
      } 
    }
  }
  if (p0) {
    var dx = (p0[0] * this.scale_[0] + this.margin_.left) / this.ratio;
    this._drawAt(p0, dx);
    return p0[3];
  }
  return null;
};

/** Get the point at a given time on the profil
 * @param { number } time time at which to show the point
 * @return { ol.coordinates } current point
 */
ol.control.Profil_beam.prototype.pointAtTime = function(time) {
  var i, p;
  // Look for closest the point
  for (i=1; p=this.tab_[i]; i++) {
    var t = p[3][3];
    if (t >= time) {
      // Previous one ?
      var pt = this.tab_[i-1][3];
      if ((pt[3]+t)/2 < time) return pt;
      else return p;
    }
  }
  return this.tab_[this.tab_.length-1][3];
};

/** Mouse move over canvas
*/
ol.control.Profil_beam.prototype.onMove = function(e) {
  if (!this.tab_.length) return;
  var box_canvas = this.canvas_.getBoundingClientRect();
  var pos = {
    top: box_canvas.top + window.pageYOffset - document.documentElement.clientTop,
    left: box_canvas.left + window.pageXOffset - document.documentElement.clientLeft
  };
  var dx = e.pageX -pos.left;
  var dy = e.pageY -pos.top;
  var ratio = this.ratio;
  if (dx>this.margin_.left/ratio && dx<(this.canvas_.width-this.margin_.right)/ratio
    && dy>this.margin_.top/ratio && dy<(this.canvas_.height-this.margin_.bottom)/ratio) {
    var d = (dx*ratio-this.margin_.left)/this.scale_[0];
    var p0 = this.tab_[0];
    for (var i=1, p; p=this.tab_[i]; i++) {
      if (p[0]>=d) {
        if (d < (p[0]+p0[0])/2) p = p0;
        break;
      }
    }
    this._drawAt(p, dx);
    this.dispatchEvent({ type:'over', click:e.type=="click", coord: p[3], time: p[2], distance: p[0] });
  } else {
    if (this.bar_.parentElement.classList.contains("over")) {
      this._drawAt();
      this.dispatchEvent({ type:'out' });
    }
  }
};

/** Show panel
* @api stable
*/
ol.control.Profil_beam.prototype.show = function() {
  this.element.classList.remove("ol-collapsed");
  this.dispatchEvent({ type:'show', show: true });
};
/** Hide panel
* @api stable
*/
ol.control.Profil_beam.prototype.hide = function() {
  this.element.classList.add("ol-collapsed");
  this.dispatchEvent({ type:'show', show: false });
};
/** Toggle panel
* @api stable
*/
ol.control.Profil_beam.prototype.toggle = function() {
  this.element.classList.toggle("ol-collapsed");
  var b = this.element.classList.contains("ol-collapsed");
  this.dispatchEvent({ type:'show', show: !b });
}
/** Is panel visible
*/
ol.control.Profil_beam.prototype.isShown = function() {
  return (!this.element.classList.contains("ol-collapsed"));
}

/**
 * Set the geometry to draw the profil.
 * @param {ol.Feature|ol.geom.Geometry} f the feature.
 * @param {Object=} options
 *  @param {ol.ProjectionLike} options.projection feature projection, default projection of the map
 *  @param {string} options.zunit 'm' or 'km', default m
 *  @param {string} options.unit 'm' or 'km', default km
 *  @param {Number|undefined} options.zmin default 0
 *  @param {Number|undefined} options.zmax default max Z of the feature
 *  @param {Number|undefined} options.graduation z graduation default 100
 *  @param {integer|undefined} options.amplitude amplitude of the altitude, default zmax-zmin
 * @api stable
 */
ol.control.Profil_beam.prototype.setGeometry = function(g, options) {
  if (!options) options = {};
  if (g instanceof ol.Feature) g = g.getGeometry();
  var canvas = this.canvas_;
  var ctx = canvas.getContext('2d');
  var w = canvas.width;
  var h = canvas.height;
  ctx.setTransform(1, 0, 0, 1, 0, 0);
  ctx.clearRect(0,0, w, h);
  // No Z
  if (!/Z/.test(g.getLayout())) return;
  // No time
  //if(/M/.test(g.getLayout())) this.element.querySelector(".time").parentElement.style.display = 'block';
  //else this.element.querySelector(".time").parentElement.style.display = 'none';
  // Coords
  var c = g.getCoordinates();
  switch (g.getType()) {
    case "LineString": break;
    case "MultiLineString": c = c[0]; break;
    default: return;
  }
  // Distance beetween 2 coords
  var proj = options.projection || this.getMap().getView().getProjection();
  function dist2d(p1,p2) {
    return ol.sphere.getDistance(
      ol.proj.transform(p1, proj, 'EPSG:4326'),
      ol.proj.transform(p2, proj, 'EPSG:4326')
    );
  }
  function getTime(t0, t1) {
    if (!t0 || !t1) return "-"
    var dt = (t1-t0) / 60; // mn
    var ti = Math.trunc(dt/60);
    var mn = Math.trunc(dt-ti*60);
    return ti+"h"+(mn<10?"0":"")+mn+"mn";
  }
  // Margin
  ctx.setTransform(1, 0, 0, 1, this.margin_.left, h-this.margin_.bottom);
  var ratio = this.ratio;
  w -= this.margin_.right + this.margin_.left;
  h -= this.margin_.top + this.margin_.bottom;
  // Draw axes
  ctx.strokeStyle = this._style.getText().getFill().getColor() || '#000';
  ctx.lineWidth = 0.5*ratio;
  ctx.beginPath();
  ctx.moveTo(0,0); ctx.lineTo(0,-h);
  ctx.moveTo(0,0); ctx.lineTo(w, 0);
  ctx.stroke();
  // Calculate [distance, altitude, time, point] for each points
  var zmin=Infinity, zmax=-Infinity;
  var i, p, z, ti, t = this.tab_ = [];
  var d = 0.0;
  var distance = 0.0;
  var climb = 0.0;
  var prevHeight;
  var grad = 0.0;
  var effort = 0.0;
  for (i=0, p; p=c[i]; i++) {
    z = p[2];
    if (z<zmin) zmin=z;
    if (z>zmax) zmax=z;
    ti = getTime(c[0][3], p[3]);
    if(i != 0){
        d = dist2d(c[i-1], p);
	distance += d;
	
        if(z >= prevHeight){
	    climb += (z - prevHeight);
	}
	if(d > 0){
        	grad = (z - prevHeight) / d;

		if(grad > -0.02){
		    effort += (d * (grad + 0.02));
		    //console.log("Effort: " + effort + " D: " + d + " Grad: " + grad);
		    //effort += 1;
		}
	}
    }
    prevHeight = z;
    
    t.push ([distance, z, ti, p, grad]);
  }
  // Info
  this.element.querySelector(".info_totalDistance").textContent = (distance/1000).toFixed(1)+this.info.distanceUnitsKM;
  this.element.querySelector(".info_climb").textContent = climb.toFixed(1) + "m";
  this.element.querySelector(".info_effort").textContent = effort.toFixed(1) + " tcEnergy";

  //this.element.querySelector(".track-info .time").textContent = ti;
  // Set graduation
  var grad = options.graduation || 100;
  while (true) {
    zmax = Math.ceil(zmax/grad)*grad;
    zmin = Math.floor(zmin/grad)*grad;
    var nbgrad = (zmax-zmin)/grad;
    if (h/nbgrad < 15*ratio) {
      grad *= 2;
    }
    else break;
  }
  // Set amplitude
  if (typeof(options.zmin)=='number' && zmin > options.zmin) zmin = options.zmin;
  if (typeof(options.zmax)=='number' && zmax < options.zmax) zmax = options.zmax;
  var amplitude = options.amplitude;
  if (amplitude) {
    zmax = Math.max (zmin + amplitude, zmax);
  }
  // Scales lines
  var scx = w/distance;
  var scy = -h/(zmax-zmin);
  var dy = this.dy_ = -zmin*scy;
  this.scale_ = [scx,scy];
  // Draw Path
  ctx.beginPath();
  for (i=0; p=t[i]; i++) {
    if (i==0) ctx.moveTo(p[0]*scx,p[1]*scy+dy);
    else ctx.lineTo(p[0]*scx,p[1]*scy+dy);
  }
  if (this._style.getStroke()) {
    ctx.strokeStyle = this._style.getStroke().getColor() || '#000';
    ctx.lineWidth = this._style.getStroke().getWidth() * ratio;
    ctx.setLineDash([]);
    ctx.stroke();
  }
  // Fill path
  if (this._style.getFill()) {
    ctx.fillStyle = this._style.getFill().getColor() || '#000';
    ctx.Style = this._style.getFill().getColor() || '#000';
    ctx.lineTo(t[t.length-1][0]*scx, 0);
    ctx.lineTo(t[0][0]*scx, 0);
    ctx.fill();
  }
  // Draw
  ctx.font = (10*ratio)+'px arial';
  ctx.textAlign = 'right';
  ctx.textBaseline = 'middle';
  ctx.fillStyle = this._style.getText().getFill().getColor() || '#000';
  // Scale Z
  ctx.beginPath();
  for (i=zmin; i<=zmax; i+=grad) {
    if (options.zunit!="km") ctx.fillText(i, -4*ratio, i*scy+dy);
    else ctx.fillText((i/1000).toFixed(1), -4*ratio, i*scy+dy);
    ctx.moveTo (-2*ratio, i*scy+dy);
    if (i!=0) ctx.lineTo (d*scx, i*scy+dy);
    else ctx.lineTo (0, i*scy+dy);
  }
  // Scale X
  ctx.textAlign = "center";
  ctx.textBaseline = "top";
  ctx.setLineDash([ratio,3*ratio]);
  var unit = options.unit ||"km";
  var step;
  if (distance>1000) {
    step = Math.round(distance/1000)*100;
    if (step > 1000) step = Math.ceil(step/1000)*1000;
  } else {
    unit = "m";
    if (distance>100) step = Math.round(distance/100)*10;
    else if (distance>10) step = Math.round(distance/10);
    else if (distance>1) step = Math.round(distance)/10;
    else step = distance;
  }
  for (i=0; i<=distance; i+=step) {
    var txt = (unit=="m") ? i : (i/1000);
    //if (i+step>d) txt += " "+ (options.zunits || "km");
    ctx.fillText(Math.round(txt*10)/10, i*scx, 4*ratio);
    ctx.moveTo (i*scx, 2*ratio); ctx.lineTo (i*scx, 0);
  }
  ctx.font = (12*ratio)+"px arial";
  ctx.fillText(this.info.xtitle.replace("(km)","("+unit+")"), w/2, 18*ratio);
  ctx.save();
  ctx.rotate(-Math.PI/2);
  ctx.fillText(this.info.ytitle, h/2, -this.margin_.left);
  ctx.restore();
  ctx.stroke();
};
/** Get profil image
* @param {string|undefined} type image format or 'canvas' to get the canvas image, default image/png.
* @param {Number|undefined} encoderOptions between 0 and 1 indicating image quality image/jpeg or image/webp, default 0.92.
* @return {string} requested data uri
* @api stable
*/
ol.control.Profil_beam.prototype.getImage = function(type, encoderOptions) {
  if (type==="canvas") return this.canvas_;
  return this.canvas_.toDataURL(type, encoderOptions);
};


class MapViewTest {
	constructor(eMap, lat, long, zoom){
		// This function displays an OpenCycleMap using the location
		this.eMap = eMap;

		this.map = new ol.Map({target: eMap,
			controls: ol.control.defaults().extend([
			    new ol.control.FullScreen()
			]),
		});

		//this.map.addLayer(new ol.layer.Tile({source:new ol.source.OSM({url: 'https://{a-c}.tile.opencyclemap.org/cycle/{z}/{x}/{y}.png'})}));
		this.map.addLayer(new ol.layer.Tile({source:new ol.source.OSM({url: 'https://{a-c}.tile.thunderforest.com/cycle/{z}/{x}/{y}.png?apikey=68bd4027130d4a278aa7b0426bf3c069'})}));

		this.map.setView(new ol.View({
			center: ol.proj.fromLonLat([long, lat]),
			zoom: zoom
		}));

		setTimeout(createObjDelegate(this, this.update), 2000);
	}
	
	update(){
		//alert("update");
		//this.map.getView().fit(this.source.getExtent(), { size: this.map.getSize(), padding: [20, 20, 20, 20]});
		this.map.updateSize();
		
		console.log("Div size: " + document.getElementById(this.eMap).offsetWidth);
		console.log("MapSize: " + this.map.getSize());
	}
}

class BHtmlNode {
	// HTML element creation and functionality
	constructor(tag, args = "", text = ""){
		this.ele = document.createElement(tag);
		var arr = args.match(/[^\s="']+|"([^"]*)"|'([^']*)'/g);
		for(var i = 0; i < arr.length; i += 2){
			//console.log("AddAttribute: " + arr[i] + " = " +  arr[i + 1].replace(/"/g, '').replace(/'/g, ''));
			this.ele.setAttribute(arr[i], arr[i + 1].replace(/"/g, '').replace(/'/g, ''));
		}
		this.ele.innerHTML = text;
	}
	
	appendNode(tag, args = "", text = ""){
		node = new BHtmlNode(atg, args, text);
		this.ele.appendChild(node.ele);
		return this;
	}
	
	setInnerHtml(str){
		this.ele.innerHTML = str;

		var arr = this.ele.getElementsByTagName('script');
		for(var i = 0; i < arr.length ; i++){
			var ele = arr[i], parent = ele.parentElement, d = document.createElement('script');
			parent.insertBefore(d, ele);
			parent.removeChild(ele);
			d.async = ele.async;
			//d.src = ele.src;
			d.textContent = ele.textContent;
		}
	}
}

function formPost(form, dialog){
	//console.log("formPost:" + form.action);
	
	fetch(form.action, {
	    method: "post",
	    body: new FormData(form),
	})
	.then(res => res.text())
	.then(res => {
		//console.log("Response: " + res);
		dialog.finish();
	})
	.catch(function (error) {
		console.log(error)
	});
	
	return false;
}

class MapDialogForm {
	constructor(map, formUrl){
		//console.log("DialogForm::constructor");
		this.map = map;
		this.formUrl = formUrl;
		//this.popup = new ol.Overlay.Popup();
		//this.map.addOverlay(this.popup);

		this.div = new BHtmlNode("div", "class='Dialog'");
		document.getElementById(this.map.eMap).appendChild(this.div.ele);

		this.div.ele.style.top = ((window.innerHeight/2) - (this.div.ele.offsetHeight/2))+'px';
		this.div.ele.style.left = ((window.innerWidth/2) - (this.div.ele.offsetWidth/2))+'px';
		this.show();
	}

	show(){
		var ele = this.div.ele;
		var dialog = this;
		fetch(this.formUrl)
			.then(res => res.text())
			.then(res => {
				this.div.setInnerHtml(res);
				var forms = ele.getElementsByTagName("form");
				forms[0].onsubmit = function(){return formPost(forms[0], dialog)};

				var cancels = ele.querySelectorAll("*[name='cancel']");
				cancels[0].onclick = createObjDelegate(this, this.close);
			})
			.catch(err => console.error(err));

		this.div.ele.style.display = "block";
	}

	close(){
		//console.log("DialogForm::close");
		this.div.ele.remove();
	}

	finish(){
		this.close();
		this.map.placesRedraw();
	}

}

var placeCategories = ["food/cafe", "food/pub", "food/shop", "food/bakery", "food/gardenCentre", "food/farmshop", "food/icecream"];
placeCategories = placeCategories.concat(["interest/castle", "interest/house", "interest/church", "interest/museum", "interest/historic", "interest/garden", "interest/beach", "interest/picnic", "interest/visitorCentre", "interest/vintageRailway", "interest/misc"]);
placeCategories = placeCategories.concat(["children/farmpark", "children/misc"]);
placeCategories = placeCategories.concat(["stay/hotel", "stay/campsite", "stay/campsiteWild", "stay/BandB", "stay/hostel", "stay/campingBarn"]);
placeCategories = placeCategories.concat(["bike/shop", "bike/pump", "bike/parking", "bike/hire", "bike/misc"]);
placeCategories = placeCategories.concat(["travel/carParking", "travel/railStation", "travel/busStation", "travel/ferry"]);
placeCategories = placeCategories.concat(["misc/summit", "misc/viewpoint", "misc/medical", "misc/toilets"]);
placeCategories = placeCategories.concat(["ride/start", "ride/finish"]);
	

class MapRide {
	// This class displays a ride on the map either as an icon or a full GPX route
	constructor(mapView, rideId, gpxUrl, lat = 0, long = 0, attr = null){
		//console.log("MapRide::constructor: ", rideId, gpxUrl);
		this.mapView = mapView;
		this.map = mapView.map;
		this.rideId = rideId;
		this.gpxUrl = gpxUrl;
		this.lat = lat;
		this.long = long;
		this.attr = attr;

		this.mode = 0;
		this.source = null;
		this.layer = null;
		this.point = null;
	}
	
	setRideMode(mode){
		this.mode = mode;
	}
	
	showRoute(){
		//console.log("MapRide::showRoute: full: " + this.full + " gpxUrl: " + this.gpxUrl);
		if(this.mode){
			this.source = new ol.source.Vector({url: this.gpxUrl + "?map=1", format: new ol.format.GPX()});
			this.layer = new ol.layer.Vector({
				source: this.source,
				style: this.mapView.styleGpx,
				zIndex: 20,
			});

			this.map.addLayer(this.layer);
			this.source.once("change", createObjDelegate(this, this.update));

			//console.log("showRoute: position: " + this.lat + "," + this.long +", name: " + this.attr.title);
			var ride = new ol.Feature({
				geometry: new ol.geom.Point(ol.proj.fromLonLat([this.long, this.lat])),
				name: this.attr ? this.attr.title : "Ride",
				id: this.rideId,
				category: "ride/start",
				group: this.attr ? this.attr.group : "",
			});

			//console.log("Category2: " + out[p].category);
			ride.setStyle(this.mapView.icons["ride/start"][0]);

			this.mapView.ridesLayerSource.addFeature(ride);

		}
		else {
			//console.log("showRoute: position: " + this.lat + "," + this.long +", name: " + this.attr.title);
			var ride = new ol.Feature({
				geometry: new ol.geom.Point(ol.proj.fromLonLat([this.long, this.lat])),
				name: this.attr ? this.attr.title : "Ride",
				id: this.rideId,
				category: "ride/start",
				group: this.attr ? this.attr.group : "",
			});

			//console.log("Category2: " + out[p].category);
			ride.setStyle(this.mapView.icons["ride/start"][0]);

			this.mapView.ridesLayerSource.addFeature(ride);
		}
	}
	
	gpxFeature(){
		return this.source.getFeatures()[0];
	}
	
	startPoint(){
		return this.gpxFeature().getGeometry().getCoordinates()[0][0].slice(0, 2);
	}

	// Draw a point on the gpxride when mouse fly over
	drawPoint(e){
		if(this.point){
			if(e.type == "over"){
				// Show point at coord
				this.point.setGeometry(new ol.geom.Point(e.coord));
				this.point.setStyle(null);
			}
			else {
				// hide point
				this.point.setStyle([]);
			}
		}
	}
	
	hover(e){
		//console.log("hover: setStyle");

		// Point on the line
		var c = this.gpxFeature().getGeometry().getClosestPoint(e.coordinate);
		this.drawPoint({ type: "over", coord: c });
		
		this.layer.setStyle(this.mapView.styleGpxSelected);
	}

	hoverLeave(e){
		//console.log("hoverLeave: setStyle");
		this.layer.setStyle(this.mapView.styleGpx);
		if(this.point){
			this.point.setStyle([]);
		}
	}

	update(e){
		//console.log("MapRide::update: ", this.source.getExtent());

		if(this.source.getState() === 'ready'){
			if(0){
				this.point = new ol.Feature(new ol.geom.Point(this.startPoint()));
			}
			else {
				this.point = new ol.Feature({
					geometry: new ol.geom.Point(this.startPoint()),
					name: this.attr ? this.attr.title : "Ride",
					id: this.rideId,
					category: "ride/start",
					group: this.attr ? this.attr.group : "",
				});
			}


			this.point.setStyle([]);
			this.source.addFeature(this.point);

			this.mapView.rideAdded(this);
		}
	}
}

class MapButton extends ol.control.Control {
	constructor(name, title, x, y){
		super({ element: document.createElement('div')});

		this.button = document.createElement("button");
		this.button.innerHTML = name;

		this.element.className = "ol-unselectable ol-control";
		this.element.style = "top: " + y + "; left: " + x;
		this.element.title = title;
		this.element.appendChild(this.button);

		//this.button.addEventListener("click", createObjDelegate(this, this.handleClick), false);
	}

	addEventListener(event, func, useCapture){
		this.button.addEventListener(event, func, useCapture);
	}
		
	handleClick(){
		alert("Called");
	}
}

class MapView {
	// This class displays an OpenCycleMap using the location or the gpx file if given
	constructor(eMap, eProfile, lat, long, zoom, showPlaces, showRides){
		this.eMap = eMap;
		this.eProfile = eProfile;
		this.position = [long, lat];
		this.zoom = zoom;
		this.showPlaces = showPlaces;
		this.showRides = showRides;
		this.rideMode = 0;
		
		this.rides = [];
		this.ride = null;
		this.rideId = 1;
		this.categories = placeCategories;
		this.hoverPlace = null;
		this.satLayer = null;
		
		this.styleGpx = new ol.style.Style({
			stroke: new ol.style.Stroke({
				color: 'rgba(255, 0, 255, 1)',
				width: 4
			}),
			image: new ol.style.RegularShape({
				radius: 10,
				radius2: 5,
				points: 5,
				fill: new ol.style.Fill({ color: 'blue' })
			}),
			fill: new ol.style.Fill({
				color: [0,0,255,0.3]
			})
		});
		this.styleGpxSelected = new ol.style.Style({
			stroke: new ol.style.Stroke({
				color: 'rgba(255, 0, 0, 1)',
				width: 4
			}),
			image: new ol.style.RegularShape({
				radius: 10,
				radius2: 5,
				points: 5,
				fill: new ol.style.Fill({ color: 'blue' })
			}),
			fill: new ol.style.Fill({
				color: [0,0,255,0.3]
			})
		});

		
		var satButton = new MapButton("A", "Aerial View", "0.5em", "6em");
		satButton.addEventListener("click", createObjDelegate(this, this.satButtonClicked), false);		

		this.map = new ol.Map({target: eMap,
			controls: ol.control.defaults().extend([
				new ol.control.FullScreen(),
				satButton,
			]),
		});
		
		// Show on map over
		this.hover = new ol.interaction.Hover({ cursor: "pointer", hitTolerance:10 });
		this.map.addInteraction(this.hover);
		this.hover.on("hover", createObjDelegate(this, this.gpxHover));
		this.hover.on("leave", createObjDelegate(this, this.gpxHoverLeave));

		if(this.eProfile != ""){
			// Profil outside the map
			this.profile = new ol.control.Profil_beam({
				target: this.eProfile,
				style: new ol.style.Style({
					fill: new ol.style.Fill({ color: '#ccc' })
				}),
				width: 400, height: isMobile?100:200
			});

			this.map.addControl(this.profile);

			// Draw the point on hover
			this.profile.on(["over","out"], createObjDelegate(this, this.profileHover));
			this.hover.on("leave", createObjDelegate(this, this.profileHoverLeave));
		}
		else {
			this.profile = null;
		}
	
		//this.map.addLayer(new ol.layer.Tile({source:new ol.source.OSM({url: 'https://{a-c}.tile.opencyclemap.org/cycle/{z}/{x}/{y}.png'})}));
		this.map.addLayer(new ol.layer.Tile({zIndex: 5, source:new ol.source.OSM({url: 'https://{a-c}.tile.thunderforest.com/cycle/{z}/{x}/{y}.png?apikey=68bd4027130d4a278aa7b0426bf3c069'})}));

		// Places layer
		this.placesLayerSource = new ol.source.Vector({});
		//this.placesLayer = new ol.layer.Vector({ source: this.placesLayerSource, zIndex: 1, minZoom: 9.75});
		this.placesLayer = new ol.layer.Vector({ source: this.placesLayerSource, zIndex: 10, minZoom: this.showPlaces});
		this.map.addLayer(this.placesLayer);
		
		// Rides layer
		this.ridesLayerSource = new ol.source.Vector({});
		this.ridesLayer = new ol.layer.Vector({ source: this.ridesLayerSource, zIndex: 15});
		this.map.addLayer(this.ridesLayer);

		this.icons = [];
		for(var i = 0; i < this.categories.length; i++){
			//console.log("AddIcon: '" + this.categories[i] + "'");
			this.icons[this.categories[i]] = this.iconCreate(this.categories[i]);
		}

		this.iconStyleInterest = new ol.style.Style({
			image: new ol.style.Icon({
				anchor: [12.0, 34.0],
				anchorXUnits: 'pixels',
				anchorYUnits: 'pixels',
				src: 'icons/maps/Place-interest.svg',
			}),
		});
		
		this.popup = new ol.Overlay.Popup();
		this.map.addOverlay(this.popup);

		this.map.on('singleclick', createObjDelegate(this, this.singleClick));
		
		this.map.setView(new ol.View({
			center: ol.proj.fromLonLat(this.position),
			zoom: zoom
		}));

		if(showPlaces)
			this.placesRedraw();

		if(showRides)
			this.ridesRedraw();
		
		window.addEventListener("resize", createObjDelegate(this, this.profileUpdate), false);
		//setTimeout(createObjDelegate(this, this.updateSize, 1), 1000);
		//this.profileUpdate();
		//setTimeout(function(){this.map.updateSize();}, 200);
	}
	
	setRideMode(mode){
		this.rideMode = mode;
	}
	
	iconCreate(category){
		var filename = "icons/maps/Place-";

		category = category.split("/");
		if(category.length == 1)
			filename += category[0] + ".svg";
		else
			filename += category[0] + "-" + category[1] + ".svg";

		//console.log("IconFilename: '" + filename + "'");
		var icon1 = new ol.style.Style({
			image: new ol.style.Icon({
				anchor: [12.0, 34.0],
				anchorXUnits: 'pixels',
				anchorYUnits: 'pixels',
				//imageSize: [48, 72],
				//size: [24, 72],
				src: filename,
				//src: 'icons/maps/PlaceFood.svg',
			}),
		});
		var icon2 = new ol.style.Style({
			image: new ol.style.Icon({
				anchor: [12.0, 34.0],
				anchorXUnits: 'pixels',
				anchorYUnits: 'pixels',
				//imageSize: [48, 72],
				//size: [24, 72],
				src: filename,
				//src: 'icons/maps/PlaceFood.svg',
				scale: 2.0,
			}),
		});
		return [icon1, icon2];
	}
	
	addRide(rideId, gpxUrl, lat, long, attr = null){
		var ride = new MapRide(this, rideId, gpxUrl, lat, long, attr);
		ride.setRideMode(this.rideMode);
		this.rides.push(ride);

		ride.showRoute();
	}
	
	rideAdded(ride){
		//console.log("MapView: showRides: " + this.showRides + " ride: " + this.ride);
		if(this.ride == null){
			this.selectRide(ride);
			if(!this.showRides){
				this.mapScaleToContent();
			}
		}
	}
	
	mapScaleToContent(){
		if(this.ride){
			this.profileUpdate();
			this.map.getView().fit(this.ride.source.getExtent(), { size: this.map.getSize(), padding: [30, 20, 20, 20]});
		}
		this.map.updateSize();
	}
	
	selectRide(ride){
		if(this.ride)
			this.ride.drawPoint({});

		this.ride = ride;

		if(this.ride && this.profile){
			this.profileUpdate();
		}
	}
	
	setFeatureBold(feature, bold){
		feature.setStyle(this.icons[feature.getProperties().category][bold]);
	}

	satButtonClicked(e){
		//alert("satButtonClicked");
		if(this.satLayer){
			if(0){
				this.map.removeLayer(this.satLayer);
				delete this.satLayer;
				this.satLayer = null
			}
			this.satLayer.setVisible(!this.satLayer.getVisible());
		}
		else {
			//alert("draw sat");
			this.satLayer = new ol.layer.Tile({
				source: new ol.source.XYZ({
					url: "https://server.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer/tile/{z}/{y}/{x}",
					maxZoom: 19,
				}),
				zIndex: 6,
				//opacity: 0.5
			});
			this.map.addLayer(this.satLayer);
		}
	}
	
	gpxHover(e){
		//console.log("MapView::gpxHover: Event: ", e);
		if((e.layer == this.placesLayer) || (e.layer == this.ridesLayer)){
			if(this.hoverPlace)
				this.setFeatureBold(this.hoverPlace, 0);
			this.setFeatureBold(e.feature, 1);
			this.hoverPlace = e.feature;
		}
		else {
			var rideNew = null;
			for(var i = 0; i < this.rides.length; i++){
				if(this.rides[i].layer == e.layer){
					rideNew = this.rides[i];
					break;
				}
			}

			if(rideNew && (this.ride != rideNew)){
				if(this.ride)
					this.ride.hoverLeave(e);
				this.selectRide(rideNew);
			}

			if(this.ride)
				this.ride.hover(e);			// Show on route
			if(this.profile)
				this.profile.showAt(e.coordinate);	// Show on profile
		}
	}

	gpxHoverLeave(e){
		//console.log("MapView::gpxHoverLeave: Event: ", e);
		if(this.hoverPlace){
			this.setFeatureBold(this.hoverPlace, 0);
			this.hoverPlace = null
		}
		if(this.ride)
			this.ride.hoverLeave(e);
	}

	profileHover(e){
		if(this.ride)
			this.ride.drawPoint(e);
	}
	
	profileHoverLeave(e){
		if(this.profile)
			this.profile.showAt();
		if(this.ride)
			this.ride.drawPoint({});
	}

	singleClick(event){
		if(event.originalEvent.shiftKey){
			var coord = ol.proj.transform(event.coordinate, 'EPSG:3857', 'EPSG:4326');
			var prettyCoord = ol.coordinate.toStringHDMS(coord, 2);
			var url = "/rides?cmd=mapPlaceAdd&longitude=" + coord[0] + "&latitude=" + coord[1];

			this.popup.show(event.coordinate, "<div><h2>Coordinates</h2><p>" + coord[1] + ", " + coord[0] + "</p><p>" + prettyCoord + "</p></div>");
			this.div = new MapDialogForm(this, url);
		}
		else {
			if(this.map.hasFeatureAtPixel(event.pixel) === true) {
				var coordinate = event.coordinate;
				var feature = this.map.forEachFeatureAtPixel(event.pixel, function(feature, layer){ return feature; });
				var html = "";
				
				//console.log("singleClick: ", feature);
				if(feature.getProperties().category == "ride/start"){
					html = "<a href='/rides?cmd=view&id=" + feature.getProperties().id + "' target='PlaceView'>" + feature.getProperties().name + "</a>";
					html += "<div>" + feature.getProperties().group + "</div>";
					html += "<div>" + feature.getProperties().category + "</div>";
				}
				else {
					html = "<a href='/rides?cmd=placeView&id=" + feature.getProperties().id + "' target='PlaceView'>" + feature.getProperties().name + "</a>";
					html += "<div>" + feature.getProperties().category + "</div>";
					html += "<div>" + feature.getProperties().description + "</div>";
					if(feature.getProperties().url != "")
						html += "<div><a target='_blank' href='" + feature.getProperties().url + "'>" + feature.getProperties().url + "</a></div>";
				}

				this.popup.show(event.coordinate, html);
				//content.innerHTML = "<a href='/rides?cmd=placeView&id=" + feature.getProperties().id + "' target='_blank'>" + feature.getProperties().name + "</a>";
				//overlay.setPosition(coordinate);
			} else {
				//closer.blur();
				this.popup.setPosition(undefined);
				if(this.div)
					this.div.close();
			}
		}
	}

	update(){	
		console.log("MapView::update: ", this.source.getExtent());
		this.map.getView().fit(this.source.getExtent(), { size: this.map.getSize(), padding: [30, 20, 20, 20]});
	}

	profileUpdate(){
		// Resize the canvas plot
		if(this.profile){
			var ep = this.profile.canvas_.parentElement;
			this.profile.canvas_.width = ep.clientWidth;
			if(this.ride){
				this.profile.setGeometry(this.ride.gpxFeature(), {
					//graduation:250,
					//amplitude:500,
					//zmin:0
				});
			}
		}
		this.map.updateSize();
	}

	ridesRedraw(){
		//console.log("ridesRedraw: lat: " + this.position[1] + " long: " + this.position[0]);
		this.placesLayerSource.clear();
		this.popup.setPosition(undefined);
		
		fetch("/rides?cmd=mapRideList&rideId=" + this.rideId + "&lat=" + this.position[1] + "&long=" + this.position[0])
			.then(res => res.json())
			.then((out) => {
				//console.log("ridesRedraw: ", out);
				for(var p in out){
					//console.log("ridesRedraw: ",  out[p].title, out[p].longitude, out[p].latitude);
					if(1){
						this.addRide(out[p].id, out[p].gpxurl, out[p].latitude, out[p].longitude, out[p]);
					}
					else {
						var ride = new ol.Feature({
							geometry: new ol.geom.Point(ol.proj.fromLonLat([out[p].longitude, out[p].latitude])),
							name: out[p].title,
							id: out[p].id,
							category: out[p].category,
							group: out[p].group,
						});

						//console.log("Category2: " + out[p].category);
						ride.setStyle(this.icons["ride/start"][0]);

						this.ridesLayerSource.addFeature(ride);
					}
				}
			})
			.catch(err => console.error(err));
	}

	placesRedraw(){
		//console.log("placesRedraw: lat: " + this.position[1] + " long: " + this.position[0]);
		this.placesLayerSource.clear();
		this.popup.setPosition(undefined);
		
		fetch("/rides?cmd=mapPlaceList&rideId=" + this.rideId + "&lat=" + this.position[1] + "&long=" + this.position[0])
			.then(res => res.json())
			.then((out) => {
				//console.log("placeMapList: ", out);
				for(var p in out){
					//console.log("place: ",  out[p].category, out[p].title, out[p].longitude, out[p].latitude);
					
					var place = new ol.Feature({
						geometry: new ol.geom.Point(ol.proj.fromLonLat([out[p].longitude, out[p].latitude])),
						name: out[p].title,
						id: out[p].id,
						category: out[p].category,
						description: out[p].description,
						url: out[p].url,
					});

					if(out[p].category in this.icons){
						//console.log("Category1: " + out[p].category);
						place.setStyle(this.icons[out[p].category][0]);
					}
					else {
						//console.log("Category2: " + out[p].category);
						place.setStyle(this.iconStyleInterest);
					}

					if(out[p].category == "ride/start"){
					}
					else {
						this.placesLayerSource.addFeature(place);
					}
				}
			})
			.catch(err => console.error(err));
	}
}
